"""
UNet structure models,
mainly for ECG wave delineation

References
----------
1. Moskalenko, Viktor, Nikolai Zolotykh, and Grigory Osipov. "Deep Learning for ECG Segmentation." International Conference on Neuroinformatics. Springer, Cham, 2019.
2. https://github.com/milesial/Pytorch-UNet/

"""

import textwrap
import warnings
from copy import deepcopy
from typing import Optional, Sequence, Union, List

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import (
    Conv_Bn_Activation,
    DownSample,
    MultiConv,
)
from ...model_configs import ECG_UNET_VANILLA_CONFIG
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    CkptMixin,
    SizeMixin,
    compute_deconv_output_shape,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "ECG_UNET",
]


class DoubleConv(MultiConv):
    """Buildings blocks for UNet.

    2 convolutions (conv --> conv) with the same number of channels.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the last convolutional layer.
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size).
    subsample_lengths : int or Sequence[int], default 1
        Subsample length(s) (stride(s)) of the convolutions.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    out_activation : bool, default True
        If True, the last mini-block of :class:`Conv_Bn_Activation`
        will have activation as in `config`; otherwise, no activation.
    mid_channels : int, optional
        Number of channels produced by the first convolutional layer,
        defaults to `out_channels`.
    config : dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers.

    """

    __name__ = "DoubleConv"

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_lengths: Union[Sequence[int], int],
        subsample_lengths: Union[Sequence[int], int] = 1,
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        out_activation: bool = True,
        mid_channels: Optional[int] = None,
        **config,
    ) -> None:
        _mid_channels = mid_channels if mid_channels is not None else out_channels
        _out_channels = [_mid_channels, out_channels]

        super().__init__(
            in_channels=in_channels,
            out_channels=_out_channels,
            filter_lengths=filter_lengths,
            subsample_lengths=subsample_lengths,
            groups=groups,
            dropouts=dropouts,
            out_activation=out_activation,
            **config,
        )


class DownDoubleConv(nn.Sequential, SizeMixin):
    """Downsampling block for the U-Net architecture.

    Downscaling with maxpool then double conv
    down sample (maxpool) --> double conv (conv --> conv)

    Channels are increased after down sampling.

    Parameters
    ----------
    down_scale : int
        Down sampling scale.
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the last convolutional layer.
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size).
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    mid_channels : int, optional
        Number of channels produced by the first convolutional layer,
        defaults to `out_channels`.
    mode : str, default "max"
        Mode for down sampling,
        can be one of {:class:`DownSample`.__MODES__}.
    config : dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers.

    """

    __name__ = "DownDoubleConv"
    __MODES__ = deepcopy(DownSample.__MODES__)

    def __init__(
        self,
        down_scale: int,
        in_channels: int,
        out_channels: int,
        filter_lengths: Union[Sequence[int], int],
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        mid_channels: Optional[int] = None,
        mode: str = "max",
        **config,
    ) -> None:
        super().__init__()
        self.__mode = mode.lower()
        assert self.__mode in self.__MODES__
        self.__down_scale = down_scale
        self.__in_channels = in_channels
        self.__mid_channels = mid_channels if mid_channels is not None else out_channels
        self.__out_channels = out_channels
        self.config = CFG(deepcopy(config))

        self.add_module(
            "down_sample",
            DownSample(
                down_scale=self.__down_scale,
                in_channels=self.__in_channels,
                norm=False,
                mode=mode,
            ),
        )
        self.add_module(
            "double_conv",
            DoubleConv(
                in_channels=self.__in_channels,
                out_channels=self.__out_channels,
                filter_lengths=filter_lengths,
                subsample_lengths=1,
                groups=groups,
                dropouts=dropouts,
                mid_channels=self.__mid_channels,
                **(self.config),
            ),
        )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the down sampling block.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        out = super().forward(input)
        return out

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the down sampling block."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class UpDoubleConv(nn.Module, SizeMixin):
    """Upsampling block of the U-Net architecture.

    Upscaling then double conv, with input of corr. down layer concatenated
    up sampling --> conv (conv --> conv)
        ^
        |
    extra input

    Channels are shrinked after up sampling.

    Parameters
    ----------
    up_scale : int
        Scale of up sampling.
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the convolutional layers.
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size) of the convolutional layers.
    deconv_filter_length : int, optional
        Length(s) of the filters (kernel size) of the
        deconvolutional upsampling layer, used only when `mode` is "deconv".
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
        Not used currently.
    deconv_groups : int, default 1
        Connection pattern (of channels) of the deconvolutional upsampling layer,
        used only when `mode` is "deconv".
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    mode : str, default "deconv"
        Mode for up sampling, can be one of {:class:`UpSample`.__MODES__}.
    mid_channels : int, optional
        Number of channels produced by the first deconvolutional layer,
        defaults to `out_channels`.
    config : dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the deconvolutional layers.

    """

    __name__ = "UpDoubleConv"
    __MODES__ = [
        "nearest",
        "linear",
        "area",
        "deconv",
    ]

    def __init__(
        self,
        up_scale: int,
        in_channels: int,
        out_channels: int,
        filter_lengths: Union[Sequence[int], int],
        deconv_filter_length: Optional[int] = None,
        groups: int = 1,
        deconv_groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        mode: str = "deconv",
        mid_channels: Optional[int] = None,
        **config,
    ) -> None:
        super().__init__()
        self.__up_scale = up_scale
        self.__in_channels = in_channels
        self.__mid_channels = (
            mid_channels if mid_channels is not None else in_channels // 2
        )
        self.__out_channels = out_channels
        self.__deconv_filter_length = deconv_filter_length
        self.__mode = mode.lower()
        assert self.__mode in self.__MODES__
        self.config = CFG(deepcopy(config))

        # the following has to be checked
        # if bilinear, use the normal convolutions to reduce the number of channels
        if self.__mode == "deconv":
            self.__deconv_padding = max(
                0, (self.__deconv_filter_length - self.__up_scale) // 2
            )
            self.up = nn.ConvTranspose1d(
                in_channels=self.__in_channels,
                out_channels=self.__in_channels,
                kernel_size=self.__deconv_filter_length,
                stride=self.__up_scale,
                padding=self.__deconv_padding,
                groups=deconv_groups,
            )
        else:
            self.up = nn.Upsample(
                scale_factor=self.__up_scale,
                mode=mode,
            )
        self.conv = DoubleConv(
            in_channels=self.__in_channels + self.__in_channels // 2,
            out_channels=self.__out_channels,
            filter_lengths=filter_lengths,
            subsample_lengths=1,
            groups=groups,
            dropouts=dropouts,
            **(self.config),
        )

    def forward(self, input: Tensor, down_output: Tensor) -> Tensor:
        """Forward pass of the up sampling block.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor from the previous layer,
            of shape ``(batch_size, n_channels, seq_len)``.
        down_output : torch.Tensor
            Input tensor of the last layer of corr. down sampling block,
            of shape ``(batch_size, n_channels', seq_len')``.

        Returns
        -------
        output : torch.Tensor
            Output tensor of the up sampling block,
            of shape ``(batch_size, n_channels'', seq_len')``.

        """
        output = self.up(input)

        diff_sig_len = down_output.shape[-1] - output.shape[-1]
        output = F.pad(output, [diff_sig_len // 2, diff_sig_len - diff_sig_len // 2])

        # TODO: consider the case `groups` > 1 when concatenating
        output = torch.cat(
            [down_output, output], dim=1
        )  # concate along the channel axis
        output = self.conv(output)

        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the up sampling block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the up sampling block.

        """
        _sep_len = seq_len
        if self.__mode == "deconv":
            output_shape = compute_deconv_output_shape(
                input_shape=[batch_size, self.__in_channels, _sep_len],
                num_filters=self.__in_channels,
                kernel_size=self.__deconv_filter_length,
                stride=self.__up_scale,
                padding=self.__deconv_padding,
            )
        else:
            output_shape = [batch_size, self.__in_channels, self.__up_scale * _sep_len]
        _, _, _seq_len = output_shape
        output_shape = self.conv.compute_output_shape(_seq_len, batch_size)
        return output_shape


class ECG_UNET(nn.Module, CkptMixin, SizeMixin, CitationMixin):
    """U-Net for (multi-lead) ECG wave delineation.

    The U-Net is a fully convolutional network originally
    proposed for biomedical image segmentation [1]_.
    This architecture is applied to ECG wave delineation in [2]_.
    This implementation is based on an open-source implementation
    on GitHub [3]_.

    Parameters
    ----------
    classes : Sequence[str]
        List of names of the classes.
    n_leads : int
        Number of input leads (number of input channels).
    config : CFG, optional,
        Other hyper-parameters, including kernel sizes, etc.
        Refer to the corresponding config file.

    References
    ----------
    .. [1] Ronneberger, Olaf, Philipp Fischer, and Thomas Brox.
           "U-net: Convolutional networks for biomedical image segmentation."
           International Conference on Medical image computing and computer-assisted intervention. Springer, 2015.
    .. [2] Moskalenko, Viktor, Nikolai Zolotykh, and Grigory Osipov.
           "Deep Learning for ECG Segmentation." International Conference on Neuroinformatics. Springer, Cham, 2019.
    .. [3] https://github.com/milesial/Pytorch-UNet/

    """

    __name__ = "ECG_UNET"

    def __init__(
        self,
        classes: Sequence[str],
        n_leads: int,
        config: Optional[CFG] = None,
    ) -> None:
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)  # final out_channels
        self.__out_channels = self.n_classes
        self.__in_channels = n_leads
        self.config = deepcopy(ECG_UNET_VANILLA_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})

        self.init_conv = DoubleConv(
            in_channels=self.__in_channels,
            out_channels=self.config.init_num_filters,
            filter_lengths=self.config.init_filter_length,
            subsample_lengths=1,
            groups=self.config.groups,
            batch_norm=self.config.batch_norm,
            activation=self.config.activation,
            kw_activation=self.config.kw_activation,
            kernel_initializer=self.config.kernel_initializer,
            kw_initializer=self.config.kw_initializer,
        )

        self.down_blocks = nn.ModuleDict()
        in_channels = self.config.init_num_filters
        for idx in range(self.config.down_up_block_num):
            self.down_blocks[f"down_{idx}"] = DownDoubleConv(
                down_scale=self.config.down_scales[idx],
                in_channels=in_channels,
                out_channels=self.config.down_num_filters[idx],
                filter_lengths=self.config.down_filter_lengths[idx],
                groups=self.config.groups,
                mode=self.config.down_mode,
                **(self.config.down_block),
            )
            in_channels = self.config.down_num_filters[idx]

        self.up_blocks = nn.ModuleDict()
        in_channels = self.config.down_num_filters[-1]
        for idx in range(self.config.down_up_block_num):
            self.up_blocks[f"up_{idx}"] = UpDoubleConv(
                up_scale=self.config.up_scales[idx],
                in_channels=in_channels,
                out_channels=self.config.up_num_filters[idx],
                filter_lengths=self.config.up_conv_filter_lengths[idx],
                deconv_filter_length=self.config.up_deconv_filter_lengths[idx],
                groups=self.config.groups,
                mode=self.config.up_mode,
                **(self.config.up_block),
            )
            in_channels = self.config.up_num_filters[idx]

        self.out_conv = Conv_Bn_Activation(
            in_channels=self.config.up_num_filters[-1],
            out_channels=self.__out_channels,
            kernel_size=self.config.out_filter_length,
            stride=1,
            groups=self.config.groups,
            norm=self.config.get("out_norm", self.config.get("out_batch_norm")),
            activation=None,
            kernel_initializer=self.config.kernel_initializer,
            kw_initializer=self.config.kw_initializer,
        )

        # for inference
        # if background counted in `classes`, use softmax
        # otherwise use sigmoid
        self.softmax = nn.Softmax(-1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        to_concat = [self.init_conv(input)]
        for idx in range(self.config.down_up_block_num):
            to_concat.append(self.down_blocks[f"down_{idx}"](to_concat[-1]))
        up_input = to_concat[-1]
        to_concat = to_concat[-2::-1]
        for idx in range(self.config.down_up_block_num):
            up_output = self.up_blocks[f"up_{idx}"](up_input, to_concat[idx])
            up_input = up_output
        output = self.out_conv(up_output)

        # to keep in accordance with other models
        # (batch_size, channels, seq_len) --> (batch_size, seq_len, channels)
        output = output.permute(0, 2, 1)

        # TODO: consider adding CRF at the tail to make final prediction

        return output

    @torch.no_grad()
    def inference(self, input: Tensor, bin_pred_thr: float = 0.5) -> Tensor:
        """Method for making inference on a single input."""
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            The length of the input signal tensor.
        batch_size : int, optional
            The batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the model.

        """
        output_shape = (batch_size, seq_len, self.n_classes)
        return output_shape

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1007/978-3-030-30425-6_29"]))



"""
configurations for signal preprocess, feature extraction, training, etc.
along with some constants

"Brady", "LAD", "RAD", "PR", "LQRSV" are treated exceptionally, as special classes
"""

from copy import deepcopy
from typing import List

from ....cfg import CFG, DEFAULTS
from ...aux_data.cinc2021_aux_data import get_class_weight

__all__ = [
    "CINC2021TrainCfg",
]


# settings from official repo
twelve_leads = (
    "I",
    "II",
    "III",
    "aVR",
    "aVL",
    "aVF",
    "V1",
    "V2",
    "V3",
    "V4",
    "V5",
    "V6",
)
six_leads = ("I", "II", "III", "aVR", "aVL", "aVF")
four_leads = ("I", "II", "III", "V2")
three_leads = ("I", "II", "V2")
two_leads = ("I", "II")


# special classes using special detectors
# _SPECIAL_CLASSES = ["Brady", "LAD", "RAD", "PR", "LQRSV"]
_SPECIAL_CLASSES = []
_NAME = "cinc2021"


def _assign_classes(cfg: CFG, special_classes: List[str]) -> None:
    """ """
    cfg.special_classes = deepcopy(special_classes)
    cfg.tranche_class_weights = CFG(
        {
            t: get_class_weight(
                t,
                exclude_classes=cfg.special_classes,
                scored_only=True,
                threshold=20,
                min_weight=cfg.min_class_weight,
            )
            for t in [
                "A",
                "B",
                "AB",
                "E",
                "F",
                "G",
            ]
        }
    )
    cfg.tranche_classes = CFG(
        {t: sorted(list(t_cw.keys())) for t, t_cw in cfg.tranche_class_weights.items()}
    )

    cfg.class_weights = get_class_weight(
        tranches="ABEFG",
        exclude_classes=cfg.special_classes,
        scored_only=True,
        threshold=20,
        min_weight=cfg.min_class_weight,
    )
    cfg.classes = sorted(list(cfg.class_weights.keys()))


# training configurations for machine learning and deep learning

CINC2021TrainCfg = CFG()
CINC2021TrainCfg.torch_dtype = DEFAULTS.DTYPE.TORCH
CINC2021TrainCfg.np_dtype = DEFAULTS.DTYPE.NP

# configs of files
CINC2021TrainCfg.db_dir = None
CINC2021TrainCfg.log_dir = DEFAULTS.log_dir / _NAME
CINC2021TrainCfg.log_dir.mkdir(parents=True, exist_ok=True)
CINC2021TrainCfg.checkpoints = DEFAULTS.checkpoints / _NAME
CINC2021TrainCfg.checkpoints.mkdir(parents=True, exist_ok=True)
CINC2021TrainCfg.model_dir = DEFAULTS.model_dir / _NAME
CINC2021TrainCfg.model_dir.mkdir(parents=True, exist_ok=True)

CINC2021TrainCfg.final_model_name = None
CINC2021TrainCfg.keep_checkpoint_max = 20

CINC2021TrainCfg.leads = deepcopy(twelve_leads)

# configs of training data
CINC2021TrainCfg.fs = 500
CINC2021TrainCfg.data_format = "channel_first"

CINC2021TrainCfg.train_ratio = 0.8
CINC2021TrainCfg.min_class_weight = 0.5
CINC2021TrainCfg.tranches_for_training = ""  # one of "", "AB", "E", "F", "G"

# assign classes, class weights, tranche classes, etc.
_assign_classes(CINC2021TrainCfg, _SPECIAL_CLASSES)

# configs of signal preprocessing
CINC2021TrainCfg.normalize = CFG(
    method="z-score",
    mean=0.0,
    std=1.0,
)
# frequency band of the filter to apply, should be chosen very carefully
CINC2021TrainCfg.bandpass = None
# CINC2021TrainCfg.bandpass = CFG(
#     lowcut=0.5,
#     highcut=60,
# )

# configs of data aumentation
# CINC2021TrainCfg.label_smooth = CFG(
#     prob=0.8,
#     smoothing=0.1,
# )
CINC2021TrainCfg.label_smooth = False
CINC2021TrainCfg.random_masking = False
CINC2021TrainCfg.stretch_compress = False  # stretch or compress in time axis
CINC2021TrainCfg.mixup = CFG(
    prob=0.6,
    alpha=0.3,
)

# configs of training epochs, batch, etc.
CINC2021TrainCfg.n_epochs = 50
# TODO: automatic adjust batch size according to GPU capacity
# https://stackoverflow.com/questions/45132809/how-to-select-batch-size-automatically-to-fit-gpu
CINC2021TrainCfg.batch_size = 64
# CINC2021TrainCfg.max_batches = 500500

# configs of optimizers and lr_schedulers
CINC2021TrainCfg.optimizer = "adamw_amsgrad"  # "sgd", "adam", "adamw"
CINC2021TrainCfg.momentum = 0.949  # default values for corresponding PyTorch optimizers
CINC2021TrainCfg.betas = (
    0.9,
    0.999,
)  # default values for corresponding PyTorch optimizers
CINC2021TrainCfg.decay = 1e-2  # default values for corresponding PyTorch optimizers

CINC2021TrainCfg.learning_rate = 1e-4  # 1e-3
CINC2021TrainCfg.lr = CINC2021TrainCfg.learning_rate

CINC2021TrainCfg.lr_scheduler = (
    "one_cycle"  # "one_cycle", "plateau", "burn_in", "step", None
)
CINC2021TrainCfg.lr_step_size = 50
CINC2021TrainCfg.lr_gamma = 0.1
CINC2021TrainCfg.max_lr = 2e-3  # for "one_cycle" scheduler, to adjust via expriments

CINC2021TrainCfg.burn_in = 400
CINC2021TrainCfg.steps = [5000, 10000]

CINC2021TrainCfg.early_stopping = CFG()  # early stopping according to challenge metric
CINC2021TrainCfg.early_stopping.min_delta = 0.001  # should be non-negative
CINC2021TrainCfg.early_stopping.patience = 10

# configs of loss function
# CINC2021TrainCfg.loss = "BCEWithLogitsLoss"
# CINC2021TrainCfg.loss = "BCEWithLogitsWithClassWeightLoss"
CINC2021TrainCfg.loss = "AsymmetricLoss"  # "FocalLoss"
CINC2021TrainCfg.loss_kw = CFG(gamma_pos=0, gamma_neg=0.2, implementation="deep-psp")
CINC2021TrainCfg.flooding_level = (
    0.0  # flooding performed if positive, typically 0.45-0.55 for cinc2021?
)

CINC2021TrainCfg.monitor = "challenge_metric"

CINC2021TrainCfg.log_step = 20
CINC2021TrainCfg.eval_every = 20

# configs of model selection
# "resnet_nature_comm_se", "multi_scopic_leadwise", "vgg16", "vgg16_leadwise",
CINC2021TrainCfg.cnn_name = "resnet_nature_comm_bottle_neck_se"
CINC2021TrainCfg.rnn_name = "none"  # "none", "lstm"
CINC2021TrainCfg.attn_name = "none"  # "none", "se", "gc", "nl"

# configs of inputs and outputs
# almost all records have duration >= 8s, most have duration >= 10s
# use `utils.utils_signal.ensure_siglen` to ensure signal length
CINC2021TrainCfg.input_len = int(500 * 10.0)
# tolerance for records with length shorter than `CINC2021TrainCfg.input_len`
CINC2021TrainCfg.input_len_tol = int(0.2 * CINC2021TrainCfg.input_len)
CINC2021TrainCfg.sig_slice_tol = 0.4  # None, do no slicing
CINC2021TrainCfg.siglen = CINC2021TrainCfg.input_len


# constants for model inference
CINC2021TrainCfg.bin_pred_thr = 0.5
# `bin_pred_look_again_tol` is used when no prob is greater than `bin_pred_thr`,
# then the prediction would be the one with the highest prob.,
# along with those with prob. no less than the highest prob. minus `bin_pred_look_again_tol`
CINC2021TrainCfg.bin_pred_look_again_tol = 0.03
CINC2021TrainCfg.bin_pred_nsr_thr = 0.1



"""
"""

from .ati_cnn import ATI_CNN_CONFIG
from .attn import global_context, non_local, squeeze_excitation, transformer
from .cnn import (  # vgg; ResNet; vanilla resnet; cpsc2018 resnet; smaller resnets; stanford resnet; ResNet Nature Communications; TResNet; cpsc2018 SOTA; multi_scopic; vanilla densenet; custom densenet; vanilla xception; custom xception; vanilla mobilenets
    cpsc_2018,
    cpsc_2018_leadwise,
    cpsc_block_basic,
    cpsc_block_mish,
    cpsc_block_swish,
    densenet_leadwise,
    densenet_vanilla,
    mobilenet_v1_vanilla,
    mobilenet_v2_vanilla,
    mobilenet_v3_small,
    multi_scopic,
    multi_scopic_block,
    multi_scopic_leadwise,
    resnet_block_basic,
    resnet_block_basic_gc,
    resnet_block_basic_nl,
    resnet_block_basic_se,
    resnet_block_stanford,
    resnet_bottle_neck,
    resnet_bottle_neck_B,
    resnet_bottle_neck_D,
    resnet_bottle_neck_gc,
    resnet_bottle_neck_nl,
    resnet_bottle_neck_se,
    resnet_cpsc2018,
    resnet_cpsc2018_leadwise,
    resnet_nature_comm,
    resnet_nature_comm_bottle_neck,
    resnet_nature_comm_bottle_neck_gc,
    resnet_nature_comm_bottle_neck_nl,
    resnet_nature_comm_bottle_neck_se,
    resnet_nature_comm_gc,
    resnet_nature_comm_nl,
    resnet_nature_comm_se,
    resnet_stanford,
    resnet_vanilla_18,
    resnet_vanilla_34,
    resnet_vanilla_50,
    resnet_vanilla_101,
    resnet_vanilla_152,
    resnet_vanilla_wide_50_2,
    resnet_vanilla_wide_101_2,
    resnetN,
    resnetNB,
    resnetNBS,
    resnetNS,
    resnext_vanilla_50_32x4d,
    resnext_vanilla_101_32x8d,
    tresnetF,
    tresnetL,
    tresnetM,
    tresnetM_V2,
    tresnetN,
    tresnetP,
    tresnetS,
    tresnetXL,
    vgg16,
    vgg16_leadwise,
    vgg_block_basic,
    vgg_block_mish,
    vgg_block_swish,
    xception_leadwise,
    xception_vanilla,
)
from .ecg_crnn import ECG_CRNN_CONFIG
from .ecg_seq_lab_net import ECG_SEQ_LAB_NET_CONFIG
from .ecg_subtract_unet import ECG_SUBTRACT_UNET_CONFIG
from .ecg_unet import ECG_UNET_VANILLA_CONFIG
from .ecg_yolo import ECG_YOLO_CONFIG
from .mlp import linear
from .rnn import attention, lstm
from .rr_lstm import RR_AF_CRF_CONFIG, RR_AF_VANILLA_CONFIG, RR_LSTM_CONFIG

__all__ = [
    # building blocks
    # CNN
    # vgg
    "vgg_block_basic",
    "vgg_block_mish",
    "vgg_block_swish",
    "vgg16",
    "vgg16_leadwise",
    # ResNet
    # resnet building blocks
    "resnet_block_basic",
    "resnet_bottle_neck",
    "resnet_bottle_neck_B",
    "resnet_bottle_neck_D",
    "resnet_block_basic_se",
    "resnet_bottle_neck_se",
    "resnet_block_basic_nl",
    "resnet_bottle_neck_nl",
    "resnet_block_basic_gc",
    "resnet_bottle_neck_gc",
    # smaller resnets
    "resnetN",
    "resnetNB",
    "resnetNS",
    "resnetNBS",
    # vanilla resnet
    "resnet_vanilla_18",
    "resnet_vanilla_34",
    "resnet_vanilla_50",
    "resnet_vanilla_101",
    "resnet_vanilla_152",
    "resnext_vanilla_50_32x4d",
    "resnext_vanilla_101_32x8d",
    "resnet_vanilla_wide_50_2",
    "resnet_vanilla_wide_101_2",
    # cpsc2018 resnet
    "resnet_cpsc2018",
    "resnet_cpsc2018_leadwise",
    # stanford resnet
    "resnet_block_stanford",
    "resnet_stanford",
    # ResNet Nature Communications
    "resnet_nature_comm",
    "resnet_nature_comm_se",
    "resnet_nature_comm_nl",
    "resnet_nature_comm_gc",
    "resnet_nature_comm_bottle_neck",
    "resnet_nature_comm_bottle_neck_se",
    "resnet_nature_comm_bottle_neck_gc",
    "resnet_nature_comm_bottle_neck_nl",
    # TresNet
    "tresnetF",
    "tresnetP",
    "tresnetN",
    "tresnetS",
    "tresnetM",
    "tresnetL",
    "tresnetXL",
    "tresnetM_V2",
    # cpsc2018 SOTA
    "cpsc_block_basic",
    "cpsc_block_mish",
    "cpsc_block_swish",
    "cpsc_2018",
    "cpsc_2018_leadwise",
    # multi_scopic
    "multi_scopic_block",
    "multi_scopic",
    "multi_scopic_leadwise",
    # vanilla densenet
    "densenet_vanilla",
    # custom densenet
    "densenet_leadwise",
    # vanilla xception
    "xception_vanilla",
    # custom xception
    "xception_leadwise",
    # vanilla mobilenets
    "mobilenet_v1_vanilla",
    "mobilenet_v2_vanilla",
    "mobilenet_v3_small",
    # RNN
    "lstm",
    "attention",
    # MLP
    "linear",
    # ATTN
    "non_local",
    "squeeze_excitation",
    "global_context",
    "transformer",
    # downstream tasks
    "ATI_CNN_CONFIG",
    "ECG_CRNN_CONFIG",
    "ECG_SEQ_LAB_NET_CONFIG",
    "ECG_SUBTRACT_UNET_CONFIG",
    "ECG_UNET_VANILLA_CONFIG",
    "ECG_YOLO_CONFIG",
    "RR_AF_CRF_CONFIG",
    "RR_AF_VANILLA_CONFIG",
    "RR_LSTM_CONFIG",
]



"""
torch_ecg.components
====================

This module contains the components for training and evaluating models.

.. contents:: torch_ecg.components
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.components

Input classes
-------------
.. autosummary::
    :toctree: generated/

    InputConfig
    WaveformInput
    FFTInput
    SpectrogramInput

Output classes
--------------
.. autosummary::
    :toctree: generated/
    :recursive:

    ClassificationOutput
    MultiLabelClassificationOutput
    SequenceTaggingOutput
    SequenceLabellingOutput
    WaveDelineationOutput
    RPeaksDetectionOutput

Loggers
-------
.. autosummary::
    :toctree: generated/
    :recursive:

    LoggerManager

Metrics
-------
.. autosummary::
    :toctree: generated/
    :recursive:

    ClassificationMetrics
    RPeaksDetectionMetrics
    WaveDelineationMetrics

Trainer
-------
.. autosummary::
    :toctree: generated/
    :recursive:

    BaseTrainer

"""

from .inputs import (
    InputConfig,
    WaveformInput,
    FFTInput,
    SpectrogramInput,
)
from .loggers import LoggerManager
from .metrics import (
    ClassificationMetrics,
    RPeaksDetectionMetrics,
    WaveDelineationMetrics,
)
from .outputs import (
    ClassificationOutput,
    MultiLabelClassificationOutput,
    SequenceTaggingOutput,
    SequenceLabellingOutput,
    WaveDelineationOutput,
    RPeaksDetectionOutput,
)
from .trainer import BaseTrainer


__all__ = [
    "InputConfig",
    "WaveformInput",
    "FFTInput",
    "SpectrogramInput",
    "LoggerManager",
    "ClassificationMetrics",
    "RPeaksDetectionMetrics",
    "WaveDelineationMetrics",
    "ClassificationOutput",
    "MultiLabelClassificationOutput",
    "SequenceTaggingOutput",
    "SequenceLabellingOutput",
    "WaveDelineationOutput",
    "RPeaksDetectionOutput",
    "BaseTrainer",
]



"""
"""

import random
import re
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import MutableMapping, Optional, Any, Union

import numpy as np
import torch

__all__ = [
    "CFG",
    "DEFAULTS",
]


_PACKAGE_CACHE = Path("~").expanduser() / ".cache" / "torch_ecg"
_PACKAGE_CACHE.mkdir(parents=True, exist_ok=True)
_DATA_CACHE = _PACKAGE_CACHE / "data"
_DATA_CACHE.mkdir(parents=True, exist_ok=True)


class CFG(dict):
    """
    This class is created in order to renew the :meth:`update` method,
    to fit the hierarchical structure of configurations.

    Examples
    --------
    >>> c = CFG(hehe={"a": 1, "b": 2})
    >>> c.update(hehe={"a": [-1]})
    >>> c
    {'hehe': {'a': [-1], 'b': 2}}
    >>> c.update(hehe={"c": -10})
    >>> c
    {'hehe': {'a': [-1], 'b': 2, 'c': -10}}
    >>> c.hehe.pop("a")
    [-1]
    >>> c
    {'hehe': {'b': 2, 'c': -10}}

    """

    __name__ = "CFG"

    def __init__(self, *args, **kwargs) -> None:
        if len(args) > 1:
            raise TypeError(f"expected at most 1 arguments, got {len(args)}")
        elif len(args) == 1:
            d = args[0]
            assert isinstance(d, MutableMapping)
        else:
            d = {}
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            try:
                setattr(self, k, v)
            except Exception:
                dict.__setitem__(self, k, v)
        # Class attributes
        exclude_fields = ["update", "pop"]
        for k in self.__class__.__dict__:
            if (
                not (k.startswith("__") and k.endswith("__"))
                and k not in exclude_fields
            ):
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x) if isinstance(x, dict) else x for x in value]
        elif isinstance(value, dict) and not isinstance(value, self.__class__):
            value = self.__class__(value)
        super().__setattr__(name, value)
        super().__setitem__(name, value)

    __setitem__ = __setattr__

    def update(self, new_cfg: Optional[MutableMapping] = None, **kwargs: Any) -> None:
        """
        The new hierarchical update method.

        Parameters
        ----------
        new_cfg : MutableMapping, optional
            The new configuration, by default None.
        **kwargs : dict, optional
            Key value pairs, by default None.

        """
        _new_cfg = new_cfg or CFG()
        if len(kwargs) > 0:  # avoid RecursionError
            _new_cfg.update(kwargs)
        for k in _new_cfg:
            if isinstance(_new_cfg[k], MutableMapping) and k in self:
                if isinstance(self[k], MutableMapping):
                    self[k].update(_new_cfg[k])
                else:  # for example, self[k] is `None`
                    self[k] = _new_cfg[k]  # deepcopy?
            else:
                try:
                    setattr(self, k, _new_cfg[k])
                except Exception:
                    dict.__setitem__(self, k, _new_cfg[k])

    def pop(self, key: str, default: Optional[Any] = None) -> Any:
        """
        The updated pop method.

        Parameters
        ----------
        key : str
            The key to pop.
        default : Any, optional
            The default value, by default None.

        """
        if key in self:
            delattr(self, key)
        return super().pop(key, default)

    def __repr__(self) -> str:
        return repr({k: v for k, v in self.items() if not callable(v)})

    def __str__(self) -> str:
        return str({k: v for k, v in self.items() if not callable(v)})


@dataclass
class DTYPE:
    """
    A dataclass to store the dtype information.

    Attributes
    ----------
    STR : str
        The string representation of the dtype.
    NP : np.dtype
        The numpy dtype.
    TORCH : torch.dtype
        The torch dtype.
    INT : int
        The int representation of the dtype, mainly used for `wfdb.rdrecord`.

    Examples
    --------
    >>> dtype = DTYPE("int16")
    >>> dtype
    DTYPE(STR='int16', NP=dtype('int16'), TORCH=torch.int16, INT=16)

    """

    STR: str
    NP: np.dtype = None
    TORCH: torch.dtype = None
    INT: int = None  # int representation of the dtype, mainly used for `wfdb.rdrecord`

    def __post_init__(self) -> None:
        """check consistency"""
        if self.NP is None:
            self.NP = np.dtype(self.STR)
        if self.TORCH is None:
            self.TORCH = eval(f"torch.{self.STR}")
        if self.INT is None:
            self.INT = int(re.search("\\d+", self.STR).group(0))
        assert all(
            [
                self.NP == getattr(np, self.STR),
                self.TORCH == getattr(torch, self.STR),
                self.INT == int(re.search("\\d+", self.STR).group(0)),
            ]
        ), "inconsistent dtype"

    @property
    def PRECISION(self) -> int:
        return self.INT


FLOAT16 = DTYPE("float16")
FLOAT32 = DTYPE("float32")
FLOAT64 = DTYPE("float64")
INT8 = DTYPE("int8")
INT16 = DTYPE("int16")
INT32 = DTYPE("int32")
INT64 = DTYPE("int64")


DEFAULTS = CFG()

DEFAULTS.log_dir = _PACKAGE_CACHE / "log"
DEFAULTS.checkpoints = _PACKAGE_CACHE / "checkpoints"
DEFAULTS.model_dir = _PACKAGE_CACHE / "saved_models"
DEFAULTS.working_dir = _PACKAGE_CACHE / "working_dir"
DEFAULTS.prefix = "TorchECG"

DEFAULTS.DTYPE = FLOAT32
DEFAULTS.DTYPE.TORCH = torch.float32  # torch.float64, torch.float16
DEFAULTS.str_dtype = str(DEFAULTS.DTYPE.TORCH).replace("torch.", "")
DEFAULTS.np_dtype = np.dtype(DEFAULTS.str_dtype)
DEFAULTS.dtype = DEFAULTS.DTYPE.TORCH

DEFAULTS.device = (
    torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
)

DEFAULTS.eps = 1e-7

DEFAULTS.SEED = 42
DEFAULTS.RNG = np.random.default_rng(seed=DEFAULTS.SEED)
DEFAULTS.RNG_sample = partial(DEFAULTS.RNG.choice, replace=False, shuffle=False)
DEFAULTS.RNG_randint = partial(DEFAULTS.RNG.integers, endpoint=True)


def set_seed(seed: int) -> None:
    """
    Set the seed of the random number generator.

    Parameters
    ----------
    seed : int
        The seed to be set.

    """

    DEFAULTS.SEED = seed
    DEFAULTS.RNG = np.random.default_rng(seed=seed)
    DEFAULTS.RNG_sample = partial(DEFAULTS.RNG.choice, replace=False, shuffle=False)
    DEFAULTS.RNG_randint = partial(DEFAULTS.RNG.integers, endpoint=True)

    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


DEFAULTS.set_seed = set_seed


def change_dtype(dtype: Union[str, np.dtype, torch.dtype]) -> None:
    """
    Change the dtype of the defaults.

    Parameters
    ----------
    dtype: str or numpy.dtype or torch.dtype,
        The dtype to be set.

    """
    # fmt: off
    _dtypes = [
        "float32", "float64", "float16",
        "int32", "int64", "int16", "int8", "uint8", "long",
    ]
    # fmt: on
    if isinstance(dtype, torch.dtype):
        _dtype = str(dtype).replace("torch.", "")
    elif isinstance(dtype, np.dtype):
        _dtype = str(dtype)
    elif isinstance(dtype, str):
        _dtype = dtype
    else:  # for example, dtype=np.float64
        try:
            _dtype = dtype.__name__
        except AttributeError:
            raise TypeError(
                f"`dtype` must be a str or np.dtype or torch.dtype, got {type(dtype)}"
            )
    assert _dtype in _dtypes, f"`dtype` must be one of {_dtypes}, got {_dtype}"
    DEFAULTS.DTYPE = DTYPE(_dtype)
    DEFAULTS.dtype = DEFAULTS.DTYPE.TORCH


DEFAULTS.change_dtype = change_dtype



"""
Utilities for convertions of data, labels, masks, etc.
"""

import os
import warnings
from collections import Counter
from copy import deepcopy
from dataclasses import dataclass
from numbers import Real
from pathlib import Path
from typing import (
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
from torch import Tensor, from_numpy
from torch.nn.functional import interpolate
from sklearn.utils import compute_class_weight
from wfdb import MultiRecord, Record
from wfdb.io import _header

from ..cfg import CFG, DEFAULTS


__all__ = [
    "get_mask",
    "class_weight_to_sample_weight",
    "rdheader",
    "ensure_lead_fmt",
    "ensure_siglen",
    "ECGWaveForm",
    "ECGWaveFormNames",
    "masks_to_waveforms",
    "mask_to_intervals",
    "uniform",
    "stratified_train_test_split",
    "cls_to_bin",
    "generate_weight_mask",
]


def get_mask(
    shape: Union[int, Sequence[int]],
    critical_points: np.ndarray,
    left_bias: int,
    right_bias: int,
    return_fmt: str = "mask",
) -> Union[np.ndarray, list]:
    """Get the mask around the given critical points.

    Parameters
    ----------
    shape : int or Sequence[int]
        Shape of the mask (and the original data).
    critical_points : numpy.ndarray
        Indices (of the last dimension) of the points
        around which to be masked (value 1).
    left_bias : int
        Bias to the left of the critical points for the mask,
        non-negative.
    right_bias : int
        Bias to the right of the critical points for the mask,
        non-negative.
    return_fmt : {"mask", "intervals"}
        Format of the return values, by default "mask".
        "mask" stands for the usual mask, while
        "intervals" means a list of intervals of the
        form ``[start, end]``.

    Returns
    -------
    numpy.ndarray or list
        The mask array or the list of intervals.

    Examples
    --------
    >>> mask = get_mask((12, 5000), np.arange(250, 5000 - 250, 400), 50, 50)
    >>> mask.shape
    (12, 5000)
    >>> mask.sum(axis=1)
    array([1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200])
    >>> intervals = get_mask((12, 5000), np.arange(250, 5000 - 250, 400), 50, 50, return_fmt="intervals")
    >>> intervals
    [[200, 300], [600, 700], [1000, 1100], [1400, 1500], [1800, 1900], [2200, 2300], [2600, 2700], [3000, 3100], [3400, 3500], [3800, 3900], [4200, 4300], [4600, 4700]]

    """
    if isinstance(shape, int):
        shape = (shape,)
    l_itv = [
        [max(0, cp - left_bias), min(shape[-1], cp + right_bias)]
        for cp in critical_points
    ]
    if return_fmt.lower() == "mask":
        mask = np.zeros(shape=shape, dtype=int)
        for itv in l_itv:
            mask[..., itv[0] : itv[1]] = 1
    elif return_fmt.lower() == "intervals":
        mask = l_itv
    return mask


def class_weight_to_sample_weight(
    y: np.ndarray, class_weight: Union[str, dict, List[float], np.ndarray] = "balanced"
) -> np.ndarray:
    """Transform class weight to sample weight.

    Parameters
    ----------
    y : numpy.ndarray
        The label (class) of each sample.
    class_weight : str or dict or List[float] or numpy.ndarray, default "balanced"
        The weight for each sample class.
        If is "balanced", the class weight will automatically be given by
        the inverse of the class frequency.
        If `y` is of string `dtype`, then `class_weight` should be a :class:`dict`.
        if `y` is of numeric `dtype`, and `class_weight` is array_like,
        then the labels (`y`) should be continuous and start from 0.

    Returns
    -------
    sample_weight : numpy.ndarray
        The array of sample weight.

    Examples
    --------
    >>> y = np.array([0, 0, 0, 0, 1, 1, 1, 2])
    >>> class_weight_to_sample_weight(y, class_weight="balanced").tolist()
    [0.25, 0.25, 0.25, 0.25, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 1.0]
    >>> class_weight_to_sample_weight(y, class_weight=[1, 1, 3]).tolist()
    [0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 1.0]
    >>> class_weight_to_sample_weight(y, class_weight={0: 2, 1: 1, 2: 3}).tolist()
    [0.6666666666666666, 0.6666666666666666, 0.6666666666666666, 0.6666666666666666, 0.3333333333333333, 0.3333333333333333, 0.3333333333333333, 1.0]
    >>> y = ["dog", "dog", "cat", "dog", "cattle", "cat", "dog", "cat"]
    >>> class_weight_to_sample_weight(y, class_weight={"dog": 1, "cat": 2, "cattle": 3}).tolist()
    [0.3333333333333333, 0.3333333333333333, 0.6666666666666666, 0.3333333333333333, 1.0, 0.6666666666666666, 0.3333333333333333, 0.6666666666666666]
    >>> class_weight_to_sample_weight(y, class_weight=[1, 2, 3])
    AssertionError: if `y` are of type str, then class_weight should be "balanced" or a dict

    """
    if not class_weight:
        sample_weight = np.ones_like(y, dtype=DEFAULTS.np_dtype)
        return sample_weight

    try:
        sample_weight = np.array(y.copy()).astype(int)
    except ValueError:
        sample_weight = np.array(y.copy())
        assert isinstance(class_weight, dict) or (
            isinstance(class_weight, str) and class_weight.lower() == "balanced"
        ), "if `y` are of type str, then class_weight should be \042balanced\042 or a dict"

    if isinstance(class_weight, str) and class_weight.lower() == "balanced":
        classes = np.unique(y).tolist()
        cw = compute_class_weight("balanced", classes=classes, y=y)
        sample_weight = np.vectorize(lambda s: cw[classes.index(s)])(sample_weight)
    else:
        sample_weight = np.vectorize(lambda s: class_weight[s])(sample_weight)
    sample_weight = sample_weight / np.max(sample_weight)
    return sample_weight


def rdheader(
    header_data: Union[Path, str, Sequence[str]]
) -> Union[Record, MultiRecord]:
    """Modified from `wfdb.rdheader`.

    Parameters
    ----------
    head_data : pathlib.Path or str or Sequence[str]
        Path of the .hea header file,
        or lines of the .hea header file.

    Returns
    -------
    wfdb.Record or wfdb.MultiRecord
        The record object.

    """
    if isinstance(header_data, (str, Path)):
        header_data = str(header_data)
        if not header_data.endswith(".hea"):
            _header_data = header_data + ".hea"
        else:
            _header_data = header_data
        if not os.path.isfile(_header_data):
            raise FileNotFoundError(f"file `{_header_data}` not found")
        with open(_header_data, "r") as f:
            _header_data = f.read().splitlines()
    elif isinstance(header_data, Sequence):
        _header_data = header_data
    else:
        raise TypeError(
            f"header_data must be str or sequence of str, but got {type(header_data)}"
        )
    # Read the header file. Separate comment and non-comment lines
    header_lines, comment_lines = [], []
    for line in _header_data:
        striped_line = line.strip()
        # Comment line
        if striped_line.startswith("#"):
            comment_lines.append(striped_line)
        # Non-empty non-comment line = header line.
        elif striped_line:
            # Look for a comment in the line
            ci = striped_line.find("#")
            if ci > 0:
                header_lines.append(striped_line[:ci])
                # comment on same line as header line
                comment_lines.append(striped_line[ci:])
            else:
                header_lines.append(striped_line)

    # Get fields from record line
    record_fields = _header._parse_record_line(header_lines[0])

    # Single segment header - Process signal specification lines
    if record_fields["n_seg"] is None:
        # Create a single-segment WFDB record object
        record = Record()

        # There are signals
        if len(header_lines) > 1:
            # Read the fields from the signal lines
            signal_fields = _header._parse_signal_lines(header_lines[1:])
            # Set the object's signal fields
            for field in signal_fields:
                setattr(record, field, signal_fields[field])

        # Set the object's record line fields
        for field in record_fields:
            if field == "n_seg":
                continue
            setattr(record, field, record_fields[field])
    # Multi segment header - Process segment specification lines
    else:
        # Create a multi-segment WFDB record object
        record = MultiRecord()
        # Read the fields from the segment lines
        segment_fields = _header._read_segment_lines(header_lines[1:])
        # Set the object's segment fields
        for field in segment_fields:
            setattr(record, field, segment_fields[field])
        # Set the objects' record fields
        for field in record_fields:
            setattr(record, field, record_fields[field])

        # Determine whether the record is fixed or variable
        if record.seg_len[0] == 0:
            record.layout = "variable"
        else:
            record.layout = "fixed"

    # Set the comments field
    record.comments = [line.strip(" \t#") for line in comment_lines]

    return record


def ensure_lead_fmt(
    values: np.ndarray, n_leads: int = 12, fmt: str = "lead_first"
) -> np.ndarray:
    """Ensure the multi-lead (ECG) signal to be of specified format.

    Parameters
    ----------
    values : numpy.ndarray
        Values of the `n_leads`-lead (ECG) signal.
    n_leads : int, default 12
        Number of leads of the multi-lead (ECG) signal.
    fmt : str, default "lead_first"
        Format of the output values, can be one of
        "lead_first" (alias "channel_first"),
        "lead_last" (alias "channel_last"), case insensitive.

    Returns
    -------
    numpy.ndarray
        ECG signal in the specified format.

    Examples
    --------
    >>> values = np.random.randn(5000, 12)
    >>> new_values = ensure_lead_fmt(values, fmt="lead_first")
    >>> new_values.shape
    (5000, 12)
    >>> np.allclose(values, new_values.T)
    True

    """
    dtype = values.dtype
    out_values = np.array(values, dtype=dtype)
    lead_dim = np.where(np.array(out_values.shape) == n_leads)[0].tolist()
    if not any([[0] == lead_dim or [1] == lead_dim]):
        raise ValueError(f"not valid {n_leads}-lead signal")
    lead_dim = lead_dim[0]
    if (lead_dim == 1 and fmt.lower() in ["lead_first", "channel_first"]) or (
        lead_dim == 0 and fmt.lower() in ["lead_last", "channel_last"]
    ):
        out_values = out_values.T
    elif fmt.lower() not in [
        "lead_first",
        "channel_first",
        "lead_last",
        "channel_last",
    ]:
        raise ValueError(f"not valid fmt: `{fmt}`")
    return out_values


def ensure_siglen(
    values: np.ndarray,
    siglen: int,
    fmt: str = "lead_first",
    tolerance: Optional[float] = None,
) -> np.ndarray:
    """Ensure the (ECG) signal to be of specified length.

    Strategy:

        1. If `values` has length greater than `siglen`,
           the central `siglen` samples will be adopted;
           otherwise, zero padding will be added to both sides.
        2. If `tolerance` is given, then if the length of `values` is
           longer than `siglen` by more than `tolerance` in percentage,
           the `values` will be sliced to have multiple of `siglen` samples,
           each with ``(1 - tolerance) * siglen`` overlap.

    Parameters
    ----------
    values : numpy.ndarray
        Values of the `n_leads`-lead (ECG) signal.
    siglen : int
        Length of the signal supposed to have.
    fmt : str, default "lead_first"
        Format of the input and output values, can be one of
        "lead_first" (alias "channel_first"),
        "lead_last" (alias "channel_last"), case insensitive.
    tolerance : float, optional
        Tolerance of the length of `values` to be
        longer than `siglen` in percentage.

    Returns
    -------
    numpy.ndarray
        ECG signal in the format of `fmt` and of fixed length `siglen`,
        of ``ndim=3`` if `tolerence` is given, otherwise ``ndim=2``.

    Examples
    --------
    >>> values = np.random.randn(12, 4629)
    >>> new_values = ensure_siglen(values, 5000, fmt="lead_first")
    >>> new_values.shape
    (12, 5000)
    >>> new_values = ensure_siglen(values, 4000, tolerance=0.1, fmt="lead_first")
    >>> new_values.shape
    (2, 12, 4000)
    >>> new_values = ensure_siglen(values, 4000, tolerance=0.2, fmt="lead_first")
    >>> new_values.shape
    (1, 12, 4000)

    """
    dtype = values.dtype
    if fmt.lower() in ["channel_last", "lead_last"]:
        # to lead_first
        _values = np.array(values, dtype=dtype).T
    else:
        _values = np.array(values, dtype=dtype).copy()
    original_siglen = _values.shape[1]
    n_leads = _values.shape[0]

    if tolerance is None or original_siglen <= siglen * (1 + tolerance):
        if original_siglen >= siglen:
            start = (original_siglen - siglen) // 2
            end = start + siglen
            out_values = _values[..., start:end]
        else:
            pad_len = siglen - original_siglen
            pad_left = pad_len // 2
            pad_right = pad_len - pad_left
            out_values = np.pad(
                _values, ((0, 0), (pad_left, pad_right)), "constant", constant_values=0
            ).astype(dtype)

        if fmt.lower() in ["channel_last", "lead_last"]:
            out_values = out_values.T
        if tolerance is not None:
            out_values = out_values[np.newaxis, ...]

        return out_values

    forward_len = int(round(siglen * tolerance))
    out_values = np.array(
        [
            _values[..., idx * forward_len : idx * forward_len + siglen]
            for idx in range((original_siglen - siglen) // forward_len + 1)
        ],
        dtype=dtype,
    )
    if fmt.lower() in ["channel_last", "lead_last"]:
        out_values = np.moveaxis(out_values, 1, -1)
    return out_values.astype(dtype)


@dataclass
class ECGWaveForm:
    """Dataclass for ECG waveform information.

    Attributes
    ----------
    name : str
        Name of the wave, e.g. "N", "p", "t", etc.
    onset : numbers.Real
        Onset index of the wave,
        :class:`~numpy.nan` for unknown/unannotated onset.
    offset : numbers.Real
        Offset index of the wave,
        :class:`~numpy.nan` for unknown/unannotated offset.
    peak : numbers.Real
        Peak index of the wave,
        :class:`~numpy.nan` for unknown/unannotated peak.
    duration : numbers.Real
        Suration of the wave, with units in milliseconds,
        :class:`~numpy.nan` for unknown/unannotated duration.

    TODO
    ----
    Add `fs` field to indicate the sampling rate of the waveform,
    and make `duration` a property computed from `fs`, `onset`, and `offset`.
    """

    name: str
    onset: Real
    offset: Real
    peak: Real
    duration: Real

    @property
    def duration_(self) -> Real:
        """Duration of the wave, with units in number of samples."""
        try:
            return self.offset - self.onset
        except TypeError:
            return np.nan


# ECGWaveForm = namedtuple(
#     typename="ECGWaveForm",
#     field_names=["name", "onset", "offset", "peak", "duration"],
# )

ECGWaveFormNames = [
    "pwave",
    "qrs",
    "twave",
]


def masks_to_waveforms(
    masks: np.ndarray,
    class_map: Dict[str, int],
    fs: Real,
    mask_format: str = "channel_first",
    leads: Optional[Sequence[str]] = None,
) -> Dict[str, List[ECGWaveForm]]:
    """Convert masks into lists of :class:`ECGWaveForm` for each lead.

    Parameters
    ----------
    masks : numpy.ndarray
        Wave delineation in the form of masks,
        of shape ``(n_leads, seq_len)``, or ``(seq_len,)``.
    class_map : dict
        Class map, mapping names to waves to numbers from 0 to n_classes-1,
        the keys should contain "pwave", "qrs", "twave".
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the `masks`,
        used to compute the duration of each waveform.
    mask_format : str, default "channel_first"
        Format of the mask, used only when ``masks.ndim = 2``, can be
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first"), case insensitive.
    leads : str or List[str], optional
        Names of leads corresponding to the channels of the `masks`.

    Returns
    -------
    dict
        Each item value is a list containing the :class:`ECGWaveForm`
        corr. to the lead.
        Each item key is from `leads` if `leads` is set,
        otherwise would be ``"lead_1", "lead_2", ..., "lead_n"``.

    Examples
    --------
    >>> class_map = {
    ...     "pwave": 1,
    ...     "qrs": 2,
    ...     "twave": 3,
    ... }
    >>> masks = np.zeros((2, 500), dtype=int)  # 2 leads, 5000 samples
    >>> masks[:, 100:150] = 1
    >>> masks[:, 160:205] = 2
    >>> masks[:, 250:340] = 3
    >>> waveforms = masks_to_waveforms(masks, class_map=class_map, fs=500, leads=["III", "aVR"])
    >>> waveforms["III"][0]
    ECGWaveForm(name='pwave', onset=100, offset=150, peak=nan, duration=100.0)

    """
    if masks.ndim == 1:
        _masks = masks[np.newaxis, ...]
    elif masks.ndim == 2:
        if mask_format.lower() not in [
            "channel_first",
            "lead_first",
        ]:
            _masks = masks.T
        else:
            _masks = masks.copy()
    else:
        raise ValueError(
            f"masks should be of dim 1 or 2, but got a {masks.ndim}d array"
        )

    _leads = (
        [f"lead_{idx+1}" for idx in range(_masks.shape[0])] if leads is None else leads
    )
    assert len(_leads) == _masks.shape[0]

    _class_map = CFG(deepcopy(class_map))

    waves = CFG({lead_name: [] for lead_name in _leads})
    for channel_idx, lead_name in enumerate(_leads):
        current_mask = _masks[channel_idx, ...]
        for wave_name, wave_number in _class_map.items():
            if wave_name.lower() not in ECGWaveFormNames:
                continue
            current_wave_inds = np.where(current_mask == wave_number)[0]
            if len(current_wave_inds) == 0:
                continue
            np.where(np.diff(current_wave_inds) > 1)
            split_inds = np.where(np.diff(current_wave_inds) > 1)[0].tolist()
            split_inds = sorted(split_inds + [i + 1 for i in split_inds])
            split_inds = [0] + split_inds + [len(current_wave_inds) - 1]
            for i in range(len(split_inds) // 2):
                itv_start = current_wave_inds[split_inds[2 * i]]
                itv_end = current_wave_inds[split_inds[2 * i + 1]] + 1
                w = ECGWaveForm(
                    name=wave_name.lower(),
                    onset=itv_start,
                    offset=itv_end,
                    peak=np.nan,
                    duration=1000 * (itv_end - itv_start) / fs,  # ms
                )
                waves[lead_name].append(w)
        waves[lead_name].sort(key=lambda w: w.onset)
    return waves


def mask_to_intervals(
    mask: np.ndarray,
    vals: Optional[Union[int, Sequence[int]]] = None,
    right_inclusive: bool = False,
) -> Union[list, dict]:
    """Convert a mask into a list of intervals,
    or a dict of lists of intervals.

    Parameters
    ----------
    mask : numpy.ndarray
        The (1D) mask to convert.
    vals : int or Sequence[int], optional
        Values in `mask` to obtain the intervals.
    right_inclusive : bool, default False
        If True, the intervals will be right inclusive,
        otherwise, right exclusive.

    Returns
    -------
    intervals : dict or list
        The intervals corr. to each value in `vals`
        if `vals` is `None` or of sequence type;
        or the intervals corr. to `vals` if `vals` is int.
        each interval is of the form ``[a, b]``.

    Examples
    --------
    >>> mask = np.zeros(100, dtype=int)
    >>> mask[10: 20] = 1
    >>> mask[80: 95] = 1
    >>> mask[50: 60] = 2
    >>> mask_to_intervals(mask, vals=1)
    [[10, 20], [80, 95]]
    >>> mask_to_intervals(mask, vals=[0, 2])
    {0: [[0, 10], [20, 50], [60, 80], [95, 100]], 2: [[50, 60]]}
    >>> mask_to_intervals(mask)
    {0: [[0, 10], [20, 50], [60, 80], [95, 100]], 1: [[10, 20], [80, 95]], 2: [[50, 60]]}
    >>> mask_to_intervals(mask, vals=[1, 2], right_inclusive=True)
    {1: [[10, 19], [80, 94]], 2: [[50, 59]]}

    """
    if vals is None:
        _vals = list(set(mask))
    elif isinstance(vals, int):
        _vals = [vals]
    else:
        _vals = vals
    # assert set(_vals) & set(mask) == set(_vals)
    bias = 0 if right_inclusive else 1

    intervals = {v: [] for v in _vals}
    for v in _vals:
        valid_inds = np.where(np.array(mask) == v)[0]
        if len(valid_inds) == 0:
            continue
        split_indices = np.where(np.diff(valid_inds) > 1)[0]
        split_indices = split_indices.tolist() + (split_indices + 1).tolist()
        split_indices = sorted([0] + split_indices + [len(valid_inds) - 1])
        for idx in range(len(split_indices) // 2):
            intervals[v].append(
                [
                    valid_inds[split_indices[2 * idx]],
                    valid_inds[split_indices[2 * idx + 1]] + bias,
                ]
            )

    if isinstance(vals, int):
        intervals = intervals[vals]

    return intervals


def uniform(low: Real, high: Real, num: int) -> List[float]:
    """Generate a list of numbers uniformly distributed.

    Parameters
    ----------
    low : numbers.Real
        Lower bound of the interval of the uniform distribution.
    high : numbers.Real
        Upper bound of the interval of the uniform distribution.
    num : int
        Number of random numbers to generate.

    Returns
    -------
    List[float]
        Array of randomly generated numbers with uniform distribution.

    Examples
    --------
    >>> arr = uniform(0, 1, 10)
    >>> all([0 <= x <= 1 for x in arr])
    True

    """
    arr = [DEFAULTS.RNG.uniform(low, high) for _ in range(num)]
    return arr


def stratified_train_test_split(
    df: pd.DataFrame,
    stratified_cols: Sequence[str],
    test_ratio: float = 0.2,
    reset_index: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Perform stratified train-test split on the dataframe.

    For example,
    if one has a dataframe with columns `sex`, `nationality`, etc.,
    assuming `sex` includes `male`, `female`; `nationality` includes `Chinese`, `American`,
    and sets `stratified_cols = ["sex", "nationality"]` with `test_ratio = 0.2`,
    then approximately 20% of the male and 20% of the female subjects
    will be put into the test set,
    and **at the same time**, approximately 20% of the Chinese and 20% of the Americans
    lie in the test set as well.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to be split.
    stratified_cols : Sequence[str]
        Columns to be stratified,
        assuming each column is a categorical variable.
        Each class in any of the columns will be split into
        train and test sets with an approximate ratio of `test_ratio`.
    test_ratio : float, default 0.2
        Ratio of test set to the whole dataframe.
    reset_index : bool, default True
        Whether to reset the index of the dataframes.

    Returns
    -------
    df_train: pandas.DataFrame
        The dataframe of the train set.
    df_test: pandas.DataFrame
        The dataframe of the test set.

    """
    invalid_cols = [
        col
        for col in stratified_cols
        if not all([v > 1 for v in Counter(df[col].apply(str)).values()])
    ]
    if len(invalid_cols) > 0:
        warnings.warn(
            f"invalid columns: {invalid_cols}, "
            "each of which has classes with only one member (row).",
            RuntimeWarning,
        )
    stratified_cols = [col for col in stratified_cols if col not in invalid_cols]
    # map to str to avoid incorrect comparison of nan values
    df_inspection = df[stratified_cols].copy().applymap(str)
    for item in stratified_cols:
        all_entities = df_inspection[item].unique().tolist()
        entities_dict = {e: str(i) for i, e in enumerate(all_entities)}
        df_inspection[item] = df_inspection[item].apply(lambda e: entities_dict[e])

    inspection_col_name = "Inspection" * (
        max([len(c) for c in stratified_cols]) // 10 + 1
    )
    df_inspection[inspection_col_name] = ""
    for idx, row in df_inspection.iterrows():
        cn = "-".join([row[sc] for sc in stratified_cols])
        df_inspection.loc[idx, inspection_col_name] = cn
    item_names = df_inspection[inspection_col_name].unique().tolist()
    item_indices = {
        n: df_inspection.index[df_inspection[inspection_col_name] == n].tolist()
        for n in item_names
    }
    for n in item_names:
        DEFAULTS.RNG.shuffle(item_indices[n])

    test_indices = []
    for n in item_names:
        item_test_indices = item_indices[n][: round(test_ratio * len(item_indices[n]))]
        test_indices += item_test_indices
    df_test = df.loc[df.index.isin(test_indices)]
    df_train = df.loc[~df.index.isin(test_indices)]
    if reset_index:
        df_train = df_train.reset_index(drop=True)
        df_test = df_test.reset_index(drop=True)
    return df_train, df_test


def cls_to_bin(
    cls_array: Union[np.ndarray, Tensor], num_classes: Optional[int] = None
) -> np.ndarray:
    """Convert a categorical array to a one-hot array.

    Convert a categorical (class indices) array of shape ``(n,)``
    to a one-hot (binary) array of shape ``(n, num_classes)``.

    Parameters
    ----------
    cls_array : numpy.ndarray or torch.Tensor
        Class indices array (tensor) of shape ``(num_samples,)``;
        or of shape ``(num_samples, num_samples)`` if `num_classes` is not None,
        in which case `cls_array` should be consistant with `num_classes`,
        and the function will return `cls_array` directly.
    num_classes : int, optional
        Number of classes. If not specified,
        it will be inferred from the values of `cls_array`.

    Returns
    -------
    numpy.ndarray
        Binary array of shape ``(num_samples, num_classes)``.

    Examples
    --------
    >>> cls_array = torch.randint(0, 26, size=(1000,))
    >>> bin_array = cls_to_bin(cls_array)
    >>> cls_array = np.random.randint(0, 26, size=(1000,))
    >>> bin_array = cls_to_bin(cls_array)

    """
    if isinstance(cls_array, Tensor):
        cls_array = cls_array.cpu().numpy()
    if num_classes is None:
        assert (
            cls_array.ndim == 1
        ), "`cls_array` should be 1D if num_classes is not specified"
        num_classes = cls_array.max() + 1
    if cls_array.ndim == 1:
        assert num_classes > 0 and num_classes >= cls_array.max() + 1, (
            "num_classes must be greater than 0 and greater than or equal to "
            "the max value of `cls_array` if `cls_array` is 1D and `num_classes` is specified"
        )
    if cls_array.ndim == 2 and cls_array.shape[1] == num_classes:
        bin_array = cls_array
    else:
        shape = (cls_array.shape[0], num_classes)
        bin_array = np.zeros(shape)
        for i in range(shape[0]):
            bin_array[i, cls_array[i]] = 1
    return bin_array


def generate_weight_mask(
    target_mask: np.ndarray,
    fg_weight: Real,
    fs: Real,
    reduction: Real,
    radius: Real,
    boundary_weight: Real,
    plot: bool = False,
) -> np.ndarray:
    """Generate weight mask for a binary target mask,
    accounting the foreground weight and boundary weight.

    Parameters
    ----------
    target_mask : numpy.ndarray
        The target mask, assumed to be 1D and binary.
    fg_weight: numbers.Real
        Foreground (value 1) weight, usually > 1.
    fs : numbers.Real
        Sampling frequency of the signal.
    reduction : numbers.Real
        Reduction ratio of the mask w.r.t. the signal.
    radius : numbers.Real
        Radius of the boundary, with units in seconds.
    boundary_weight : numbers.Real
        Weight for the boundaries (positions where values change)
        of the target map.
    plot : bool, default False
        If True, `target_mask` and the result `weight_mask` will be plotted.

    Returns
    -------
    weight_mask : numpy.ndarray
        Weight mask generated from `target_mask`.

    Examples
    --------
    >>> target_mask = np.zeros(50000, dtype=int)
    >>> target_mask[500:14000] = 1
    >>> target_mask[35800:44600] = 1
    >>> fg_weight = 2.0
    >>> fs = 500
    >>> reduction = 1
    >>> radius = 0.8
    >>> boundary_weight = 5.0
    >>> weight_mask = generate_weight_mask(
    ...     target_mask, fg_weight, fs, reduction, radius, boundary_weight
    ... )
    >>> weight_mask.shape
    (50000,)
    >>> reduction = 10
    >>> weight_mask = generate_weight_mask(
    ...     target_mask, fg_weight, fs, reduction, radius, boundary_weight
    ... )
    >>> weight_mask.shape
    (5000,)

    """
    assert target_mask.ndim == 1, "`target_mask` should be 1D"
    assert set(np.unique(target_mask)).issubset(
        {0, 1}
    ), "`target_mask` should be binary"
    assert (
        isinstance(reduction, Real) and reduction >= 1
    ), "`reduction` should be a real number greater than 1"
    if reduction > 1:
        # downsample the target mask
        target_mask = (
            interpolate(
                from_numpy(target_mask).unsqueeze(0).unsqueeze(0).float(),
                mode="nearest",
                scale_factor=1 / reduction,
                recompute_scale_factor=True,
            )
            .squeeze(0)
            .squeeze(0)
            .numpy()
            .astype(DEFAULTS.DTYPE.NP)
        )
    weight_mask = np.ones_like(target_mask, dtype=DEFAULTS.DTYPE.NP)
    sigma = int((radius * fs) / reduction)
    weight = np.full_like(target_mask, fg_weight) - 1
    weight_mask += (target_mask > 0.5) * weight
    border = np.where(np.diff(target_mask) != 0)[0]
    for idx in border:
        # weight = np.zeros_like(target_mask, dtype=DEFAULTS.DTYPE.NP)
        # weight[max(0, idx-sigma): (idx+sigma)] = boundary_weight
        weight = np.full_like(target_mask, boundary_weight, dtype=DEFAULTS.DTYPE.NP)
        weight = weight * np.exp(
            -np.power(np.arange(len(target_mask)) - idx, 2) / sigma**2
        )
        weight_mask += weight
    if plot:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(figsize=(12, 6))
        ax.plot(target_mask, label="target mask")
        ax.plot(weight_mask, label="weight mask")
        ax.set_xlabel("samples")
        ax.set_ylabel("weight")
        ax.legend(loc="best")
        plt.show()

    return weight_mask



# -*- coding: utf-8 -*-

import re
import warnings
from pathlib import Path
from typing import Any, List, Optional, Union, Sequence

import numpy as np
import pandas as pd
from scipy.io import loadmat

from ...cfg import DEFAULTS
from ...utils.misc import get_record_list_recursive, add_docstring
from ...utils.download import http_get
from ..base import DEFAULT_FIG_SIZE_PER_SEC, CPSCDataBase, DataBaseInfo


__all__ = [
    "CPSC2018",
    "compute_metrics",
]


_CPSC2018_INFO = DataBaseInfo(
    title="""
    The China Physiological Signal Challenge 2018:
    Automatic identification of the rhythm/morphology abnormalities in 12-lead ECGs.
    """,
    about="""
    1. training set contains 6,877 (female: 3178; male: 3699) 12 leads ECG recordings lasting from 6 s to just 60 s.
    2. ECG recordings were sampled as 500 Hz.
    3. the training data can be downloaded using links in [1]_, but the link in [2]_ is recommended. File structure will be assumed to follow [2]_.
    4. the training data are in the ``channel first`` format.
    5. types of abnormal rhythm/morphology + normal in the training set:

        +-----+-------------------------------------+-------+-------------------+
        | No. |   name                              | abbr. | number of records |
        +=====+=====================================+=======+===================+
        | 0   | Normal                              | N     | 918               |
        +-----+-------------------------------------+-------+-------------------+
        | 1   | Atrial fibrillation                 | AF    | 1098              |
        +-----+-------------------------------------+-------+-------------------+
        | 2   | First-degree atrioventricular block | I-AVB | 704               |
        +-----+-------------------------------------+-------+-------------------+
        | 3   | Left bundle brunch block            | LBBB  | 207               |
        +-----+-------------------------------------+-------+-------------------+
        | 4   | Right bundle brunch block           | RBBB  | 1695              |
        +-----+-------------------------------------+-------+-------------------+
        | 5   | Premature atrial contraction        | PAC   | 556               |
        +-----+-------------------------------------+-------+-------------------+
        | 6   | Premature ventricular contraction   | PVC   | 672               |
        +-----+-------------------------------------+-------+-------------------+
        | 7   | ST-segment depression               | STD   | 825               |
        +-----+-------------------------------------+-------+-------------------+
        | 8   | ST-segment elevated                 | STE   | 202               |
        +-----+-------------------------------------+-------+-------------------+

    6. ordering of the leads in the data of all the records are

        .. code-block:: python

            ["I", "II", "III", "aVR", "aVL", "aVF", "V1", "V2", "V3", "V4", "V5", "V6"]

    7. meanings in the .hea files: **to write**
    8. knowledge about the abnormal rhythms: ref. :meth:`get_disease_knowledge`.
    9. Challenge official website [1]_, see also [2]_.
    """,
    note="""
    1. Ages of records A0608, A1549, A1876, A2299, A5990 are "NaN".
    2. CINC2020 (ref. [2]_) released totally 3453 unused training data of CPSC2018, whose filenames start with "Q".
       These file names are not "continuous". The last record is "Q3581".
    """,
    usage=[
        "ECG arrythmia detection",
    ],
    references=[
        "http://2018.icbeb.org/",
        "https://physionetchallenges.github.io/2020/",
    ],
    doi="10.1166/jmihi.2018.2442",
)


@add_docstring(_CPSC2018_INFO.format_database_docstring(), mode="prepend")
class CPSC2018(CPSCDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "CPSC2018"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="cpsc2018",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = 500
        self.spacing = 1000 / self.fs
        self.rec_ext = "mat"
        self.ann_ext = "mat"  # the same file as the record

        self.nb_records = 6877
        self.all_leads = [
            "I",
            "II",
            "III",
            "aVR",
            "aVL",
            "aVF",
            "V1",
            "V2",
            "V3",
            "V4",
            "V5",
            "V6",
        ]
        self.all_diagnosis = [
            "N",
            "AF",
            "I-AVB",
            "LBBB",
            "RBBB",
            "PAC",
            "PVC",
            "STD",
            "STE",
        ]
        self.all_diagnosis_original = sorted(
            [
                "Normal",
                "AF",
                "I-AVB",
                "LBBB",
                "RBBB",
                "PAC",
                "PVC",
                "STD",
                "STE",
            ]
        )
        self.diagnosis_abbr_to_full = {
            "N": "Normal",
            "AF": "Atrial fibrillation",
            "I-AVB": "First-degree atrioventricular block",
            "LBBB": "Left bundle brunch block",
            "RBBB": "Right bundle brunch block",
            "PAC": "Premature atrial contraction",
            "PVC": "Premature ventricular contraction",
            "STD": "ST-segment depression",
            "STE": "ST-segment elevated",
        }
        self.diagnosis_num_to_abbr = {
            1: "N",
            2: "AF",
            3: "I-AVB",
            4: "LBBB",
            5: "RBBB",
            6: "PAC",
            7: "PVC",
            8: "STD",
            9: "STE",
        }

        self._all_records = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        self._df_records = pd.DataFrame()
        self._df_records["path"] = get_record_list_recursive(
            self.db_dir, self.rec_ext, relative=False
        )
        if self._subsample is not None:
            size = min(
                len(self._df_records),
                max(1, int(round(self._subsample * len(self._df_records)))),
            )
            self._df_records = self._df_records.sample(
                n=size, random_state=DEFAULTS.SEED, replace=False
            )
        self._df_records["path"] = self._df_records["path"].apply(lambda x: Path(x))
        self._df_records["record"] = self._df_records["path"].apply(lambda x: x.name)
        self._df_records.set_index("record", inplace=True)
        self._all_records = self._df_records.index.values.tolist()

        # find and load the annotation csv file
        ann_file = list(self.db_dir.rglob("REFERENCE.csv"))
        if len(ann_file) != 1:
            warnings.warn(
                "Annotation file not found. Please call method `_download_labels`, "
                "and call method `_ls_rec` again.",
                RuntimeWarning,
            )
            for c in ["labels_n", "labels_a", "labels_f"]:
                self._df_records.at[self._df_records.index, c] = None
        else:
            df_ann = pd.read_csv(ann_file[0])
            label_mat = df_ann[df_ann.columns[1:]].values
            mask = np.isnan(label_mat)
            label_mat = [
                row[~mask[idx]].astype(int).tolist()
                for idx, row in enumerate(label_mat)
            ]
            df_ann.loc[df_ann.index, "labels_n"] = df_ann.apply(
                lambda row: label_mat[row.name], axis=1
            )
            df_ann.loc[df_ann.index, "labels_a"] = df_ann.apply(
                lambda row: [self.diagnosis_num_to_abbr[i] for i in row.labels_n],
                axis=1,
            )
            df_ann.loc[df_ann.index, "labels_f"] = df_ann.apply(
                lambda row: [self.diagnosis_abbr_to_full[i] for i in row.labels_a],
                axis=1,
            )
            df_ann = df_ann[[df_ann.columns[0], "labels_n", "labels_a", "labels_f"]]
            df_ann.columns = ["record", "labels_n", "labels_a", "labels_f"]
            df_ann.set_index("record", inplace=True)
            # merge `df_ann` and `self._df_records`
            self._df_records = self._df_records.merge(
                df_ann, how="left", left_index=True, right_index=True
            )

    def get_subject_id(self, rec: Union[int, str]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        s2d = {
            "A": "11",
            "B": "12",
        }
        prefix = "".join(re.findall(r"[A-Z]", rec))
        n = rec.replace(prefix, "")
        sid = int(f"{s2d[prefix]}{'0'*(8-len(n))}{n}")
        return sid

    def load_data(
        self,
        rec: Union[int, str],
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        data_format="channel_first",
        units: str = "mV",
    ) -> np.ndarray:
        """Load the ECG data of a record.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or Sequence[str] or Sequence[int], optional
            The leads to load,
            None or "all" for all leads.
        data_format : str, default "channel_first"
            Format of the ECG data, "channel_last" or "channel_first" (original)
        units : str, default "mV"
            Units of the output signal, can also be "μV" (with an alias "uV"),
            case insensitive.

        Returns
        -------
        data : numpy.ndarray
            The loaded ECG data.

        """
        if isinstance(rec, int):
            rec = self[rec]
        rec_fp = self.get_absolute_path(rec, self.rec_ext)
        _leads = self._normalize_leads(leads, numeric=True)
        allowed_data_format = [
            "channel_first",
            "lead_first",
            "channel_last",
            "lead_last",
        ]
        assert (
            data_format.lower() in allowed_data_format
        ), f"`data_format` should be one of `{allowed_data_format}`, but got `{data_format}`"

        allowed_units = ["mv", "uv", "μv", "muv"]
        assert (
            units is None or units.lower() in allowed_units
        ), f"`units` should be one of `{allowed_units}` or None, but got `{units}`"

        data = loadmat(str(rec_fp))
        data = np.asarray(data["ECG"]["data"][0, 0], dtype=DEFAULTS.DTYPE.NP)[_leads, :]
        if data_format.lower() in ["channel_last", "lead_last"]:
            data = data.T
        if units.lower() == "mv" and self._auto_infer_units(data) != "mV":
            data /= 1000
        elif (
            units.lower() in ["uv", "μv", "muv"]
            and self._auto_infer_units(data) != "μV"
        ):
            data *= 1000

        return data

    def load_ann(self, rec: Union[str, int], ann_format: str = "n") -> List[str]:
        """Load labels (diagnoses or arrhythmias) of a record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        ann_format : str, default "n"
            Format of labels, one of the following (case insensitive):

                - "a", abbreviations
                - "f", full names
                - "n", numeric codes

        Returns
        -------
        labels : List[str]
            The list of labels.

        """
        if isinstance(rec, int):
            rec = self[rec]
        try:
            col = {
                "a": "labels_a",
                "f": "labels_f",
                "n": "labels_n",
            }[ann_format.lower()]
        except KeyError:
            raise ValueError(
                f"`ann_format` should be one of `['a', 'f', 'n']`, but got `{ann_format}`"
            )
        labels = self._df_records.loc[rec, col]
        return labels

    @add_docstring(load_ann.__doc__)
    def get_labels(self, rec: Union[str, int], ann_format: str = "n") -> List[str]:
        """alias of `load_ann`"""
        return self.load_ann(rec, ann_format)

    def get_subject_info(
        self, rec: Union[int, str], items: Optional[List[str]] = None
    ) -> dict:
        """Get subject information (e.g sex, age, etc.).

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.
        items : List[str], optional
            Items of the subject information (e.g. sex, age, etc.).

        Returns
        -------
        subject_info : dict
            The subject information.

        """
        if items is None or len(items) == 0:
            info_items = ["age", "sex"]
        else:
            info_items = items

        if isinstance(rec, int):
            rec = self[rec]
        rec_fp = self.get_absolute_path(rec, self.ann_ext)
        data = loadmat(str(rec_fp))["ECG"]
        subject_info = {
            "age": data["age"][0, 0][0, 0],
            "sex": data["sex"][0, 0][0],
        }

        return subject_info

    def plot(
        self,
        rec: Union[int, str],
        ticks_granularity: int = 0,
        leads: Optional[Union[str, List[str]]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot the ECG data of a record.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads : str or List[str], optional
            The leads to plot
        kwargs : dict, optional
            Auxilliary key word arguments to pass to :func:`matplotlib.pyplot.subplots`.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

        leads = self._normalize_leads(leads)

        data = self.load_data(rec, data_format="channel_first", units="μV", leads=leads)
        y_ranges = np.max(np.abs(data), axis=1) + 100

        diag = self.get_labels(rec, ann_format="f")

        nb_leads = len(leads)

        t = np.arange(data.shape[1]) / self.fs
        duration = len(t) / self.fs
        fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * duration))
        fig_sz_h = 6 * y_ranges / 1500
        fig, axes = plt.subplots(
            nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h))
        )
        for idx in range(nb_leads):
            axes[idx].plot(
                t,
                data[idx],
                color="black",
                linewidth="2.0",
                label=f"lead - {leads[idx]}",
            )
            axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            if ticks_granularity >= 1:
                axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                axes[idx].grid(
                    which="major", linestyle="-", linewidth="0.5", color="red"
                )
            if ticks_granularity >= 2:
                axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                axes[idx].grid(
                    which="minor", linestyle=":", linewidth="0.5", color="black"
                )
            axes[idx].plot([], [], " ", label=f"labels - {','.join(diag)}")
            axes[idx].legend(loc="upper left")
            axes[idx].set_xlim(t[0], t[-1])
            axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
            axes[idx].set_xlabel("Time [s]")
            axes[idx].set_ylabel("Voltage [μV]")
        plt.subplots_adjust(hspace=0.2)
        if kwargs.get("save_path", None):
            plt.savefig(kwargs["save_path"], dpi=200, bbox_inches="tight")
        else:
            plt.show()

    @property
    def url(self) -> List[str]:
        return [
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet1.zip",
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet2.zip",
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet3.zip",
        ]

    @property
    def database_info(self) -> DataBaseInfo:
        return _CPSC2018_INFO

    @property
    def webpage(self) -> str:
        return "http://2018.icbeb.org/Challenge.html"

    def _download_labels(self) -> None:
        label_url = "http://2018.icbeb.org/file/REFERENCE.csv"
        http_get(label_url, self.db_dir, extract=False)

    def download(self) -> None:
        """Download the database from :attr:`self.url`."""
        for url in self.url:
            http_get(url, self.db_dir, extract=True)
        self._download_labels()
        self._ls_rec()


def compute_metrics():
    """ """
    raise NotImplementedError
